(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Arith.
Require Import MinMax.

Require Import list_fall_exst.
  
Require Import predicate.
Require Import closure.
Require Import carrier.
Require Import graph.
Require Import maps.
Require Import finite.
Require Import enum.
 
Require Import formula.
Require Import labels.
Require Import words.
Require Import constraints.
Require Import css.
Require Import pme.

Require Import kripke.

Require Import branch.

Require Import bbi_simple.
Require Import bbi_basic.
Require Import bbi_expansion.
Require Import bbi_strategy.
Require Import bbi_oracle.
Require Import bbi_sequence. 
Require Import bbi_hintikka.

Section Limit.

  Notation eps := (nil : word label).

  Let c := idx_dmap _ idx_label.
  Let c_inj : injective c.
  Proof. apply indexation_dinj. Qed.

  Variable   istm  : nat -> stm.
  Hypothesis Histm : fair_strategy istm.

  Variable   P  : CSS_type -> Prop.
  Hypothesis Pdec : forall css, P css \/ ~ P css.
  Hypothesis HP : CSS_consistent inc P.
  Hypothesis OP : oracle P.

  Variable   n0 : nat.
  Variable   X0 : stm -> Prop.
  Variable   C0 : cst -> Prop.

  Hypothesis HXC0 : finite_CSS (X0,C0).
  Hypothesis H0   : C0 (eps ~~ eps).
  Hypothesis H1   : CSS_consistent (X0,C0).
  Hypothesis H2   : alphabet C0 inc pd_image c (fun n => n < n0).

  Hypothesis HC0  : bbi_basic_pme (pme C0).

  Let P_id X C X' C' : X ~ X' -> C ~ C' -> P (X,C) -> P (X',C').
  Proof.
    intros G0 G1 G2.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    apply O2 with (3 := G2).
    apply O1 in G2.
    generalize G2; apply CSS_id_id; auto. 
    split; simpl; auto.
    apply G0.
    apply G1.
  Qed.

  Implicit Type p : CSS_type*cst.

  Let Q i p := match p with (X,C,_) => P (X cup sg (istm i), C) end.
  
  Let Qdec i p : Q i p \/ ~ Q i p.
  Proof. destruct p as ((X,C),ce); simpl. apply Pdec. Qed.

  Let next_false (i : nat) p := match p with (X,C,_) => (X,C cup sg (eps~~eps), eps~~eps) end.

  Let next_cup i stmi :=
    let a := c (n0+2*i)   in
    let b := c (n0+2*i+1) in
    match stmi with
        | T I      # m => (empty                             ,       eps ~~ m)
        | T A *  B # m => (sg (T A # l a) cup sg (T B # l b) , l a o l b ~~ m)
        | F A -* B # m => (sg (T A # l a) cup sg (F B # l b) , l a o   m ~~ l b)
        | _            => (empty                             ,       eps ~~ eps)
    end.

  Let next_Xe i stm_i := fst (next_cup i stm_i).
  Let next_ce i stm_i := snd (next_cup i stm_i).

  Let next_true  (i : nat) p := 
    match p with 
      (X,C,_) => ((X cup sg (istm i)) cup next_Xe i (istm i), 
                   C cup sg (next_ce i (istm i)), 
                   next_ce i (istm i)) 
    end.

  Let P' i p := 
    match p with 
      (X,C,_) => C (eps~~eps) 
              /\ P (X,C) 
              /\ alphabet C inc pd_image c (fun n => n < n0+2*i) 
    end.

  Hint Extern 0 ( pd_image _ _ inc pd_image _ _ ) => apply pd_image_inc; intro; omega.

  Let Hnext_false : forall n x, P' n x -> ~ Q n x -> P' (S n) (next_false n x).
  Proof.
    intros n ((X,C),ce) [ G1 [ G2 G3 ] ] G4.
    unfold next_false.
    split; pred_solve.
    split.
    apply P_id with (3 := G2); auto.
    apply inc_transitive with (1 := proj1 (alphabet_cup _ _)); pred_solve.
    apply inc_transitive with (1 := G3); auto.
    apply inc_transitive with (1 := @alphabet_sg_eps _); auto.
  Qed.

  Ltac intro_hyp H P := 
     let H1 := fresh in
     generalize P; intros H1; 
     match goal with [ H1: ?x -> _ |- _ ] => assert x as H end; 
     [ clear H1 | apply H1 in H; clear H1 ].

  Let Hnext_true : forall n x, P' n x -> Q n x -> P' (S n) (next_true n x).
  Proof.
    intros n ((X,C),ce) [ G1 [ G2 G3 ] ] G4.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    unfold next_true.
    assert ( P' (S n) ((X cup sg (istm n)) cup empty, C cup sg (eps~~eps), next_ce n (istm n)) ) as HP'.
      split; pred_solve.
      split.
      unfold Q in G4.
      apply P_id with (3 := G4); auto.
      apply inc_transitive with (1 := proj1 (alphabet_cup _ _)); pred_solve.
      apply inc_transitive with (1 := G3); auto.      
      apply inc_transitive with (1 := @alphabet_sg_eps _); auto.
    set (m := snd (snd (istm n))).
    assert (pme C (m ~~ m)) as Hm.
      destruct HP' as [ _ [ HP' _ ] ].
      apply O1 in HP'.
      destruct (istm n) as (s,(f,m')).
      simpl in m; unfold m.
      apply pme_eps_1.
      apply (HP' s f m'); pred_solve.
    unfold m in Hm; clear m.
    destruct (istm n) as ([],([ | [] | [] A | [] A B ],m)); unfold next_cup; auto;
    destruct HP' as [HP1 [ HP2 HP3 ] ]; simpl in Hm.
    
    split.
    unfold next_ce; pred_solve.
    split.
    intro_hyp H (bbi_exp_TI (X cup sg (T I # m)) C m).
    pred_solve.
    unfold next_ce, next_cup; simpl.
    destruct O5 with (2 := H) as (t & [ Ht1 | [] ] & Ht2).
    apply O1 in HP2.
    generalize HP2; apply CSS_id_id; auto.
    apply P_id with (3 := HP2); auto.
    rewrite <- Ht1 in Ht2; auto.
    unfold next_ce, next_cup; simpl; clear HP1.
    apply inc_transitive with (2 := HP3).
    apply inc_transitive with (1 := proj1 (alphabet_cup _ _)); pred_solve.
    apply inc_transitive with (1 := proj1 (alphabet_sg _ _)); pred_solve.
    apply inc_transitive with (2 := proj2 (alphabet_cup _ _)); pred_solve.

    split. 
    pred_solve.
    intro_hyp H (bbi_exp_Tmult (X cup sg (T A * B # m)) C A B (c (n0+2*n)) (c (n0+2*n+1)) m).
    pred_solve.
    destruct O5 with (2 := H) as (t & [ Ht1 | [] ] & Ht2).

    apply O1 in HP2.
    generalize HP2; apply CSS_id_id; auto.  
    apply P_id with (3 := HP2); auto.
 
    rewrite <- Ht1 in Ht2; auto.
    split; auto.
    clear HP1.
    unfold next_ce, next_cup, snd.
    apply inc_transitive with (1 := proj1 (alphabet_cup _ _)); pred_solve.
    apply inc_transitive with (1 := G3); auto.
    apply inc_transitive with (1 := proj1 (alphabet_sg _ _)); pred_solve.
    apply inc_transitive with (1 := proj1 (carrier_app _ _ _)); pred_solve;
    apply carrier_cons_2; auto; [ exists (n0+2*n) | exists (n0+2*n+1) ]; split; auto; omega.
    apply inc_transitive with (2 := HP3).
    apply inc_transitive with (2 := proj2 (alphabet_cup _ _)); pred_solve.
  
    intros Hn; apply c_inj in Hn; omega.

    intros Hn.
    apply G3 in Hn.
    destruct Hn as [ ? [ ? Hn] ].
    apply c_inj in Hn; omega.

    intros Hn.
    apply G3 in Hn.
    destruct Hn as [ ? [ ? Hn] ].
    apply c_inj in Hn; omega.
  
    unfold next_ce, next_cup, snd.
    split.
    pred_solve.
    split.
    intro_hyp H (bbi_exp_Fwand (X cup sg (F A -* B # m)) C A B (c (n0+2*n)) (c (n0+2*n+1)) m).
    pred_solve.
    destruct O5 with (2 := H) as (t & [ Ht1 | [] ] & Ht2).

    apply O1 in HP2.
    generalize HP2; apply CSS_id_id; auto.
    apply P_id with (3 := HP2); auto.

    rewrite <- Ht1 in Ht2; auto.

    intros Hn; apply c_inj in Hn; omega.

    intros Hn.
    apply G3 in Hn.
    destruct Hn as [ ? [ ? Hn] ].
    apply c_inj in Hn; omega.

    intros Hn.
    apply G3 in Hn.
    destruct Hn as [ ? [ ? Hn] ].
    apply c_inj in Hn; omega.
    
    clear HP1.
    apply inc_transitive with (1 := proj1 (alphabet_cup _ _)); pred_solve.
    apply inc_transitive with (1 := G3), pd_image_inc; intro; omega.
    apply inc_transitive with (1 := proj1 (alphabet_sg _ _)); pred_solve.
    apply inc_transitive with (1 := proj1 (carrier_app _ _ _)); pred_solve.
    apply carrier_cons_2; auto.
    exists (n0+2*n); split; auto; omega.
    apply inc_transitive with (2 := HP3).
    apply inc_transitive with (2 := proj2 (alphabet_cup _ _)); pred_solve.
    apply carrier_cons_2; auto.
    exists (n0+2*n+1); split; auto; omega.
  Qed.
  
  Let bbi_limit := bbi_sequence _ Q (X0,C0,eps~~eps) next_false next_true.
  
  Let bbi_limit_Pall i Xi Ci ci : bbi_limit (i,(Xi,Ci,ci)) -> Ci (eps~~eps) /\ P (Xi,Ci) /\ alphabet Ci inc pd_image c (fun n => n < n0+2*i).
  Proof.
    apply bbi_seq_P with (P := P'); auto.
    repeat (split; auto).
    apply inc_transitive with (1 := H2); auto.
  Qed.

  Let bbi_limit_function : function_graph bbi_limit.
  Proof. apply bbi_seq_function; auto. Qed.

  Let bbi_limit_fun : functional_graph bbi_limit.
  Proof. apply function_graph_functional; auto. Qed.

  Let bbi_limit_prop0 i Xi Ci ci : bbi_limit (i,(Xi,Ci,ci)) -> Ci (eps~~eps).
  Proof. intro H; apply bbi_limit_Pall with (1 := H). Qed.

  Let bbi_limit_prop1 i Xi Ci ci : bbi_limit (i,(Xi,Ci,ci)) -> P(Xi,Ci).
  Proof. intro H; apply bbi_limit_Pall with (1 := H). Qed.

  Let bbi_limit_prop2 i Xi Ci ci : bbi_limit (i,(Xi,Ci,ci)) -> alphabet Ci inc pd_image c (fun n => n < n0+2*i).
  Proof. intro H; apply bbi_limit_Pall with (1 := H). Qed.

  Let bbi_limit_O : bbi_limit (0,(X0,C0,eps~~eps)).
  Proof. apply bbi_seq_O; auto. Qed.

  Let X i s := exists Xi, exists Ci, exists ci, bbi_limit (i,(Xi,Ci,ci)) /\ Xi s.
  Let C i c := exists Xi, exists Ci, exists ci, bbi_limit (i,(Xi,Ci,ci)) /\ Ci c.
  Let is_ce q := match q with (i,ce) => exists Xi, exists Ci, exists ci, bbi_limit (i,(Xi,Ci,ci)) /\ ci = ce end.

  Let X_prop i Xi Ci ci : bbi_limit (i,(Xi,Ci,ci)) -> X i ~ Xi.
  Proof.
    intros H; split.
    intros x ( ? & ? & ? & HXC & ? ).
    injection (bbi_limit_fun _ _ _ HXC H).
    intros _ _ H'; rewrite <- H'; auto.
    intros x Hx; exists Xi; exists Ci; exists ci; auto.
  Qed.

  Let C_prop i Xi Ci ci : bbi_limit (i,(Xi,Ci,ci)) -> C i ~ Ci.
  Proof.
    intros H; split.
    intros x ( ? & ? & ? & HXC & ? ).
    injection (bbi_limit_fun _ _ _ HXC H).
    intros _ H' _; rewrite <- H'; auto.
    intros x Hx; exists Xi; exists Ci; exists ci; auto.
  Qed.

  Let ce_prop i Xi Ci ci : bbi_limit (i,(Xi,Ci,ci)) -> is_ce (i,ci).
  Proof. intros; exists Xi; exists Ci; exists ci; auto. Qed.

  Let is_ce_function : function_graph is_ce.
  Proof.
    intros i.
    destruct (bbi_limit_function i) as ( ((Xi,Ci),ci) & H ).
    apply gr_dimage_sg' in H.
    exists ci.
    split.
    intros x ( y & G1 & Xi' & Ci' & ci' & G2 & G3 ).
    pred_solve.
    rewrite <- G1 in G2.
    rewrite <- G3.
    injection (bbi_limit_fun _ _ _ H G2).
    intros E; rewrite E; auto.
    apply sg_inc.
    exists i.
    split; auto.
    exists Xi; exists Ci; exists ci; auto.
  Qed.

  Let is_ce_functional : functional_graph is_ce.
  Proof. apply function_graph_functional; auto. Qed.

  Let XC_prop i : exists Xi, exists Ci, exists ci, bbi_limit (i,(Xi,Ci,ci)).
  Proof.
    destruct (bbi_limit_function i) as [ ((Xi,Ci),ci) Hi ].
    exists Xi; exists Ci; exists ci; apply gr_dimage_sg' in Hi; auto.
  Qed.

  Let X_0 : X 0 ~ X0.           Proof. apply (X_prop _ _ _ _ bbi_limit_O). Qed. 
  Let C_0 : C 0 ~ C0.           Proof. apply (C_prop _ _ _ _ bbi_limit_O). Qed.

  Let is_ce_0 ce : is_ce (0,ce) <-> ce = eps~~eps.
  Proof.
    split.
    intros (Xi & Ci & ci & G0 & G1).
    rewrite <- G1.
    injection (bbi_limit_fun _ _ _ G0 bbi_limit_O); auto.
    intros G0; rewrite G0; clear ce G0.
    exists X0; exists C0; exists (eps~~eps); auto.
  Qed.

  Let C_prop0 i : C i (eps~~eps).
  Proof.
    destruct (XC_prop i) as ( ? & ? & ? & Hi ).
    apply C_prop with (1 := Hi).
    apply bbi_limit_prop0 with (1 := Hi).
  Qed.

  Let XC_prop1 i : P (X i, C i).
  Proof.
    destruct (XC_prop i) as ( Xi & Ci & ci & Hi ).
    apply P_id with (3 := bbi_limit_prop1 _ _ _ _ Hi); auto; 
    apply id_symetric.
    apply (X_prop _ _ Ci ci); auto.
    apply (C_prop _ Xi _ ci); auto.
  Qed.
  
  Let C_prop2 i : alphabet (C i) inc pd_image c (fun n => n < n0+2*i).
  Proof.
    destruct (XC_prop i) as ( Xi & Ci & ci & Hi).
    apply inc_transitive with (2 := bbi_limit_prop2 _ _ _ _ Hi).
    apply alphabet_inc.
    apply (C_prop _ Xi _ ci); auto.
  Qed.

  Let not_in_alphabet i q : n0+2*i <= q -> ~ alphabet (pme (C i)) (c q).
  Proof.
    intros F1 F0.
    apply alphabet_pme in F0.
    apply C_prop2 in F0.
    destruct F0 as (x & F0 & F2).
    apply c_inj in F2.
    omega.
  Qed.

  Let X_prop3 i : X i inc X (S i).
  Proof.
    intros z ( Xi & Ci & ci & Hz1 & Hz2 ).
    destruct (Qdec i (Xi,Ci,ci)).

    unfold bbi_limit in Hz1. 
    apply bbi_seq_Strue in Hz1; auto.
    simpl next_true in Hz1.
    unfold next_Xe, next_ce in Hz1.
    destruct (next_cup i (istm i)) as (Xe,ce).
    exists ((Xi cup sg (istm i)) cup Xe).
    exists (Ci cup sg ce).
    exists ce.
    split; pred_solve.
    
    unfold bbi_limit in Hz1. 
    apply bbi_seq_Sfalse in Hz1; auto.
    simpl next_false in Hz1.
    exists Xi.
    exists (Ci cup sg (eps~~eps)).
    exists (eps~~eps).
    split; auto.
  Qed.

  Let C_prop3' i ce : is_ce (S i,ce) -> C (S i) ~ C i cup sg ce.
  Proof.
    intros Hce.
    destruct (bbi_limit_function i) as ( ((Xi,Ci),ci) & Hi ).
    apply gr_dimage_sg' in Hi.
    generalize (C_prop _ _ _ _ Hi).
    intros HCi.
    destruct (Qdec i (Xi,Ci,ci)).
    
    unfold bbi_limit in Hi. 
    apply bbi_seq_Strue in Hi; auto.
    simpl next_true in Hi.
    unfold next_Xe, next_ce in Hi.
    destruct (next_cup i (istm i)) as (Xe,ce').
    simpl fst in Hi.
    simpl snd in Hi.
    generalize (ce_prop _ _ _ _ Hi)
               (C_prop  _ _ _ _ Hi).
    intros Hce' Hsi.
    rewrite (is_ce_functional _ _ _ Hce Hce').
    apply id_transitive with (1 := Hsi); auto. 

    unfold bbi_limit in Hi. 
    apply bbi_seq_Sfalse in Hi; auto.
    simpl next_false in Hi.
    generalize (ce_prop _ _ _ _ Hi)
               (C_prop  _ _ _ _ Hi).
    intros Hce' Hsi.
    rewrite (is_ce_functional _ _ _ Hce Hce').
    apply id_transitive with (1 := Hsi); auto. 
  Qed.

  Let C_prop3 i : C i inc C (S i).
  Proof.
    destruct (is_ce_function (S i)) as (ce & He).
    apply gr_dimage_sg' in He.
    generalize (C_prop3' _ _ He).
    intros [ ? H' ].
    apply inc_transitive with (2 := H'); auto.
  Qed.

  Let ce_basic i ce : is_ce (S i, ce) -> bbi_basic (pme (C i)) ce.
  Proof.
    intros H.
    destruct (XC_prop i) as ( Xi & Ci & ci & Hi ).
    generalize  (X_prop  _ _ _ _ Hi)
                (C_prop  _ _ _ _ Hi).
    intros G6 G7.
    destruct (Qdec i (Xi,Ci,ci)) as [ G9 | G9 ].
    
    unfold bbi_limit in Hi. 
    apply bbi_seq_Strue in Hi; auto.
    unfold next_true, next_Xe, next_ce in Hi.
    generalize (ceq_eps (@pme_is_pme _ (C i))).
    intros G10.
    red in G9.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    apply O1 in G9.

    destruct (istm i) as [ [] [ [ | [] | [] A | [] A B ] ? ] ];
    apply ce_prop in Hi;
    rewrite (is_ce_functional _ _ _ H Hi);
    unfold next_cup, fst, snd;
    try (constructor; auto);
    try (
    apply pme_mono with (1 := proj2 G7);
    match goal with 
      [ G9: CSS (_ cup (sg (?s,(?f,_))), _) |- _ ] => apply (G9 s f)
    end;
    apply bin_union_inc_r2; auto );
    try (intros E; apply c_inj in E; omega);
    try (apply not_in_alphabet; omega).
    
    unfold bbi_limit in Hi. 
    apply bbi_seq_Sfalse in Hi; auto.
    unfold next_true, next_Xe, next_ce in Hi.
    generalize  (ce_prop  _ _ _ _ Hi).
    intros G10.
    rewrite (is_ce_functional _ _ _ H G10).
    constructor.
    apply (ceq_eps (@pme_is_pme _ _)).
  Qed.
 
  Let XC_prop_next_1 i : P (X i cup sg (istm i), C i) -> X (S i) (istm i).
  Proof.
    intros H.
    destruct (XC_prop i) as ( Xi & Ci & ci & Hi ).
    unfold bbi_limit in Hi. 
    apply bbi_seq_Strue in Hi; auto.
    simpl in Hi.
    unfold next_Xe, next_ce in Hi.
    destruct (next_cup i (istm i)) as (Xe,ce).
    exists ((Xi cup sg (istm i)) cup Xe).
    exists (Ci cup sg ce).
    exists ce.
    split; auto.
    apply bin_union_inc_r1; auto.
    apply bin_union_inc_r2; auto.
    red.
    apply P_id with (3 := H); auto.
    generalize (X_prop _ _ _ _ Hi); intro; auto.
    apply (C_prop _ _ _ _ Hi).
  Qed.
  
  Let XC_prop_next_2 i m : istm i = T I # m 
                            -> P (X i cup sg (T I # m), C i) 
                            -> X (S i) ~ X i cup (sg (T I # m))
                            /\ C (S i) ~ C i cup sg (eps ~~ m).
  Proof.
    intros G0 G1.
    destruct (XC_prop i) as ( Xi & Ci & ci & Hi ).
    generalize (X_prop _ _ _ _ Hi)
              (C_prop  _ _ _ _ Hi).
    intros G6 G7.
    unfold bbi_limit in Hi. 
    apply bbi_seq_Strue in Hi; auto.
    unfold next_true, next_Xe, next_ce in Hi.
    destruct (istm i) as [ ? [] ]; try discriminate G0.
    injection G0.
    clear G0; intros G0 G2 G3.
    rewrite G0, G2, G3 in Hi.
    simpl next_cup in Hi.
    simpl fst in Hi.
    simpl snd in Hi.
    generalize (X_prop _ _ _ _ Hi)
              (C_prop  _ _ _ _ Hi).
    intros G4 G5.
    split.
    apply id_transitive with (1 := G4); auto.
    destruct G6; split; pred_solve.
    apply id_transitive with (1 := G5); auto.
    red.
    rewrite G0.
    apply P_id with (3 := G1); auto.
  Qed.

  Let XC_prop_next_3 i A B m : istm i = T A * B # m 
                                -> P (X i cup sg (T A * B # m), C i) 
                                -> X (S i) ~ (X i cup  sg (T A * B # m)) 
                                                  cup (sg (T A # l (c (n0+2*i)))
                                                  cup  sg (T B # l (c (n0+2*i+1))))
                                /\ C (S i) ~  C i cup sg (l (c (n0+2*i)) o l (c (n0+2*i+1)) ~~ m).
  Proof.
    intros G0 G1.
    destruct (XC_prop i) as ( Xi & Ci & ci & Hi ).
    generalize (X_prop _ _ _ _ Hi)
              (C_prop  _ _ _ _ Hi).
    intros G6 G7.
    unfold bbi_limit in Hi. 
    apply bbi_seq_Strue in Hi; auto.
    unfold next_true, next_Xe, next_ce in Hi.
    destruct (istm i) as [ ? [] ]; try discriminate G0.
    injection G0.
    clear G0; intros G0 G2 G3.
    rewrite G0, G2, G3 in Hi.
    simpl next_cup in Hi.
    simpl snd in Hi.
    simpl fst in Hi.
    generalize (X_prop _ _ _ _ Hi)
              (C_prop  _ _ _ _ Hi).
    intros G4 G5.
    split.
    apply id_transitive with (1 := G4); auto.
    destruct G6; split; pred_solve.
    apply id_transitive with (1 := G5); auto.
    apply G5; pred_solve.
    red.
    rewrite G0.
    apply P_id with (3 := G1); auto.
  Qed.

  Let XC_prop_next_4 i A B m : istm i = F A -* B # m 
                                -> P (X i cup sg (F A -* B # m), C i) 
                                -> X (S i) ~ (X i cup  sg (F A -* B # m)) 
                                                  cup (sg (T A # l (c (n0+2*i)))
                                                  cup  sg (F B # l (c (n0+2*i+1))))
                                /\ C (S i) ~ C i cup sg (l (c (n0+2*i)) o m ~~ l (c (n0+2*i+1))).
  Proof.
    intros G0 G1.
    destruct (XC_prop i) as ( Xi & Ci & ci & Hi ).
    generalize (X_prop _ _ _ _ Hi)
              (C_prop  _ _ _ _ Hi).
    intros G6 G7.
    unfold bbi_limit in Hi. 
    apply bbi_seq_Strue in Hi; auto.
    unfold next_true, next_Xe, next_ce in Hi.
    destruct (istm i) as [ ? [] ]; try discriminate G0.
    injection G0.
    clear G0; intros G0 G2 G3.
    rewrite G0, G2, G3 in Hi.
    simpl next_cup in Hi.
    simpl snd in Hi.
    simpl fst in Hi.
    generalize (X_prop _ _ _ _ Hi)
              (C_prop  _ _ _ _ Hi).
    intros G4 G5.
    split.
    apply id_transitive with (1 := G4); auto.
    destruct G6; split; pred_solve.
    apply id_transitive with (1 := G5); auto.
    apply G5; pred_solve.
    red.
    rewrite G0.
    apply P_id with (3 := G1); auto.
  Qed.

  (* The sequence (X i, C i, is_ce (i,_)) is built *)

  (* Let us now build the limit of this sequence *)

  Definition lX := union X.
  Definition lC := union C.

  Theorem lC_simple : bbi_simple_pme (pme lC).
  Proof.
    set (f q := match q with (n,c) => is_ce (S n,c) end).
    apply (basic_seq_is_simple) with (C0 := C0) (ce := f); auto.
    apply bbi_basic_id.
    apply function_compose_function_graph; auto.
  Qed.

  Let Xk_inc_lX k : X k inc lX.
  Proof. apply union_inc_r. Qed.

  Let Ck_inc_lC k : C k inc lC.
  Proof. apply union_inc_r. Qed.   

  Theorem XC0_inc_lXC : CSS_leq (X0,C0) (lX,lC).
  Proof. 
    split.
    apply inc_transitive with (1 := proj2 X_0); auto.
    apply inc_transitive with (1 := proj2 C_0); auto.
  Qed.

  Let lXC_CSS : CSS (lX,lC).
  Proof.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    intros s f m H.
    apply union_sg in H.
    destruct H as [ n Hn ]. 
    apply (O1 _ (XC_prop1 n)) in Hn.
    apply pme_mono with (2 := Hn); auto.
  Qed.

  Hint Resolve XC0_inc_lXC.

  Let lXC_P : P (lX,lC).
  Proof.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    apply O3; auto.
    intros (Xf,Cf) ( Hf1 & Hf2 & Hf3 ) ( Hf4 & Hf5 ).
    destruct (finite_union_seq _ _ X_prop3 _ Hf2 Hf4) as [ nX HnX ].
    destruct (finite_union_seq _ _ C_prop3 _ Hf3 Hf5) as [ nC HnC ].
    set (n := max nX nC).
    apply (O2 _ (X n,C n)); auto.
    split; auto.
    apply inc_transitive with (1 := HnX).
    apply (seq_inc _ _ X_prop3). unfold n; apply le_max_l.
    apply inc_transitive with (1 := HnC).
    apply (seq_inc _ _ C_prop3). unfold n; apply le_max_r.
  Qed.

  Let lXC_max s f m : P (lX cup sg (s,(f,m)), lC) -> lX (s,(f,m)).
  Proof.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    intros G0.
    generalize (O1 _ G0); intros G1.
    assert (pme lC (m~~m)) as G2.
      apply (G1 s f m); pred_solve.
    apply pme_cpct in G2.
    destruct G2 as ( Cf & G2 & G3 & G4 ).
    destruct (finite_union_seq _ _ C_prop3 _ G2 G3) as [ j Hj ].
    destruct (Histm (s,(f,m)) j) as ( k & Hk1 & Hk2 ).
    red in Hk2.
    assert (pme (C k) (m~~m)) as G5.
      apply pme_mono with (R := C j).
      apply (seq_inc _ _ C_prop3); auto.
      apply pme_mono with (1 := Hj); auto.
    assert (P (X k cup sg (s,(f,m)), C k)) as G6.
      apply O2 with (3 := G0).
      apply CSS_inc_stm_l; auto.
      split; pred_solve.
    rewrite <- Hk2 in G6.
    apply XC_prop_next_1 in G6.
    rewrite Hk2 in G6.
    apply Xk_inc_lX with (1 := G6); auto.
  Qed.

  Let lXC_max_I m : lX (T I # m) -> lC (eps ~~ m).
  Proof.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    unfold lX. 
    intros H.
    apply union_sg in H.
    destruct H as [ i Hi ].
    destruct (Histm (T I # m) i) as ( k & Hk1 & Hk2 ).
    red in Hk2.
    assert (X k (T I # m)) as G0.
      apply (seq_inc _ _ X_prop3 _ _ Hk1); auto.
    assert (P (X k cup sg (istm k), C k)) as G1.
      red in O2.
      apply O2 with (3 := XC_prop1 k).
      rewrite Hk2.
      apply CSS_inc_stm_l; auto.
      apply pme_mono with (R := C i); auto.
      apply (seq_inc _ _ C_prop3); auto.
      apply (O1 _ (XC_prop1 i) _ _ _ Hi).
      split; pred_solve.
      rewrite Hk2; auto.
    apply (union_inc_r _ _ _ (S k)).
    rewrite Hk2 in G1. 
    apply XC_prop_next_2 in G1; auto.
    apply (proj2 G1); pred_solve.
  Qed.

  Let lXC_max_mult A B m : lX (T A * B # m) -> exists x, exists y, lX (T A # x) /\ lX (T B # y) /\ lC (x o y ~~ m).
  Proof.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    intros H.
    unfold lX. 
    apply union_sg in H.
    destruct H as [ i Hi ].
    destruct (Histm (T A * B # m) i) as ( k & Hk1 & Hk2 ).
    red in Hk2.
    assert (X k (T A * B # m)) as G0.
      apply (seq_inc _ _ X_prop3 _ _ Hk1); auto.
    assert (P (X k cup sg (istm k), C k)) as G1.
      red in O2.
      apply O2 with (3 := XC_prop1 k).
      rewrite Hk2.
      apply CSS_inc_stm_l; auto.
      apply (O1 _ (XC_prop1 k) _ _ _ G0).
      split; pred_solve.
      rewrite Hk2; auto.
    exists (l (c (n0+2*k))).
    exists (l (c (n0+2*k+1))).
    rewrite Hk2 in G1.
    apply XC_prop_next_3 in G1; auto.
    destruct G1 as ( G1 & G2 ).
    split.
    apply Xk_inc_lX with (k := S k), G1; pred_solve.
    split.
    apply Xk_inc_lX with (k := S k), G1; pred_solve.
    apply Ck_inc_lC with (k := S k), G2; pred_solve.
  Qed.

  Let lXC_max_wand A B m : lX (F A -* B # m) -> exists x, exists y, lX (T A # x) /\ lX (F B # y) /\ lC (x o m ~~ y).
  Proof.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    intros H.
    unfold lX. 
    apply union_sg in H.
    destruct H as [ i Hi ].
    destruct (Histm (F A -* B # m) i) as ( k & Hk1 & Hk2 ).
    red in Hk2.
    assert (X k (F A -* B # m)) as G0.
      apply (seq_inc _ _ X_prop3 _ _ Hk1); auto.
    assert (P (X k cup sg (istm k), C k)) as G1.
      red in O2.
      apply O2 with (3 := XC_prop1 k).
      rewrite Hk2.
      apply CSS_inc_stm_l; auto.
      apply (O1 _ (XC_prop1 k) _ _ _ G0).
      split; pred_solve.
      rewrite Hk2; auto.
    exists (l (c (n0+2*k))).
    exists (l (c (n0+2*k+1))).
    rewrite Hk2 in G1.
    apply XC_prop_next_4 in G1; auto.
    destruct G1 as ( G1 & G2 ).
    split.
    apply Xk_inc_lX with (k := S k), G1; pred_solve.
    split.
    apply Xk_inc_lX with (k := S k), G1; pred_solve.
    apply Ck_inc_lC with (k := S k), G2; pred_solve.
  Qed.

  Theorem lXC_Hintikka : Hintikka lX lC.
  Proof.
    destruct OP as [ O1 O2 O3 O4 O5 ].
    split; red; auto.

    generalize (O4 (lX,lC) lXC_CSS lXC_P).
    intros G0 v m n G1 G2 G3.
    apply G0.
    apply bin_inter_sg'; auto.
    constructor 1 with (1 := G1) (2 := G2); auto.

    intros A m G0.
    destruct (O5 _ _ lXC_CSS (bbi_exp_Tnot _ _ _ _ G0) lXC_P) as ( t & [ G1 | [] ] & G2 ).
    rewrite <- G1 in G2.
    apply lXC_max.
    apply O2 with (3 := G2); auto.

    intros A m G0.
    destruct (O5 _ _ lXC_CSS (bbi_exp_Fnot _ _ _ _ G0) lXC_P) as ( t & [ G1 | [] ] & G2 ).
    rewrite <- G1 in G2.
    apply lXC_max.
    apply O2 with (3 := G2); auto.

    intros A B m G0.
    destruct (O5 _ _ lXC_CSS (bbi_exp_Tand _ _ _ _ _ G0) lXC_P) as ( t & [ G1 | [] ] & G2 ).
    rewrite <- G1 in G2.
    split;
    apply lXC_max;
    apply O2 with (3 := G2); auto;
    apply O1 in G2;
    apply CSS_inc_empty_r in G2;
    try (split; auto; fail);
    apply CSS_dec_l with (2 := G2); pred_solve.
    
    intros A B m G0.
    destruct (O5 _ _ lXC_CSS (bbi_exp_Fand _ _ _ _ _ G0) lXC_P) as ( t & [ G1 | [ G1 | [] ] ] & G2 ); 
    [ left | right ]; rewrite <- G1 in G2;
    apply lXC_max;
    apply O2 with (3 := G2); auto;
    apply O1 in G2;
    apply CSS_inc_empty_r in G2;
    try (split; auto; fail);
    apply CSS_dec_l with (2 := G2); pred_solve.

    generalize (O4 (lX,lC) lXC_CSS lXC_P).
    intros G0 m G1 G2.
    apply G0.
    apply bin_inter_sg'; auto.
    constructor 2 with (1 := G1) (2 := G2); auto.
    
    intros A B m G0.
    apply lXC_max_mult in G0.
    destruct G0 as ( x & y & G1 & G2 & G3 ).
    exists x; exists y; auto.
    
    intros A B m G0 x y G1.
    destruct (O5 _ _ lXC_CSS (bbi_exp_Fmult _ _ _ _ _ _ _ G0 G1) lXC_P) as ( t & [ G2 | [ G2 | [] ] ] & G3 ); 
    [ left | right ]; rewrite <- G2 in G3;
    apply lXC_max;
    apply O2 with (3 := G3); auto;
    apply O1 in G3;
    apply CSS_inc_empty_r in G3;
    try (split; auto; fail);
    apply CSS_dec_l with (2 := G3); pred_solve.

    intros A B m G0 x y G1.
    destruct (O5 _ _ lXC_CSS (bbi_exp_Twand _ _ _ _ _ _ _ G0 G1) lXC_P) as ( t & [ G2 | [ G2 | [] ] ] & G3 ); 
    [ left | right ]; rewrite <- G2 in G3;
    apply lXC_max;
    apply O2 with (3 := G3); auto;
    apply O1 in G3;
    apply CSS_inc_empty_r in G3;
    try (split; auto; fail);
    apply CSS_dec_l with (2 := G3); pred_solve.

    intros A B m G0.
    apply lXC_max_wand in G0.
    destruct G0 as ( x & y & G1 & G2 & G3 ).
    exists x; exists y; auto.
  Qed.

End Limit.

