(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import predicate.
Require Import graph.
Require Import uniq_choice. 

Section bbi_sequence.

  Variable X : Type.

  Variable   Q    : nat -> X -> Prop.
  Hypothesis Qdec : forall n x, Q n x \/ ~ Q n x.

  Variable   P  : nat -> X -> Prop.
  Variable   x0 : X.
  Hypothesis H0 : P 0 x0.

  Variables (next_false next_true : nat -> X -> X). 
  
  Hypothesis Hnext_false : forall n x, P n x -> ~ Q n x -> P (S n) (next_false n x).
  Hypothesis Hnext_true  : forall n x, P n x ->   Q n x -> P (S n) (next_true  n x).

  Let R n c : Prop := match c with (x,y) => (   Q n x /\ y = next_true  n x )
                                         \/ ( ~ Q n x /\ y = next_false n x )
                      end.

  Let R_functional n :functional_graph (R n).
  Proof.
    intros x y z [ [H1 H2] | [H1 H2] ] [ [H3 H4] | [H3 H4] ];
    try (exfalso; tauto); rewrite H4; auto.
  Qed.

  Let R_total n : total_graph (R n).
  Proof.
    intros x;
    destruct (Qdec n x) as [ H | H ];
    [ exists (next_true n x); left
    | exists (next_false n x); right
    ]; auto.
  Qed.
    
  Let R_function n : function_graph (R n).
  Proof. apply functional_total_graph_function; auto. Qed.
   
  Definition bbi_sequence := uniq_choice x0 R.
  
  Proposition bbi_seq_function : function_graph bbi_sequence.
  Proof. apply uniq_choice_function; auto. Qed.

  Proposition bbi_seq_O : bbi_sequence (0,x0).
  Proof. apply uniq_choice_O; auto. Qed.

  Proposition bbi_seq_Sfalse n x : ~ Q n x -> bbi_sequence (n,x) -> bbi_sequence (S n, next_false n x).
  Proof.
    intros H1 H2.
    generalize (uniq_choice_S x0 R n x).
    intros H3.
    apply H3 with (2 := H2).
    right; tauto.
  Qed.

  Proposition bbi_seq_Strue n x : Q n x -> bbi_sequence (n,x) -> bbi_sequence (S n, next_true  n x).
  Proof.
    intros H1 H2.
    generalize (uniq_choice_S x0 R n x).
    intros H3.
    apply H3 with (2 := H2).
    left; tauto.
  Qed.

  Proposition bbi_seq_P n : forall x, bbi_sequence (n,x) -> P n x.
  Proof.
    generalize (function_graph_functional _ _ _ bbi_seq_function); intros H.
    induction n as [ | n IHn ]; intros x Hx.

    rewrite (H _ _ _ Hx bbi_seq_O); auto.

    destruct (bbi_seq_function n) as [ y Hy ].
    apply gr_dimage_sg' in Hy.
    generalize (IHn _ Hy); intros Hy'.
    destruct (Qdec n y) as [ HP | HP ]. 
    
    generalize (bbi_seq_Strue _ _ HP Hy).
    intros H'. 
    rewrite (H _ _ _ Hx H').
    apply Hnext_true; auto.

    generalize (bbi_seq_Sfalse _ _ HP Hy).
    intros H'. 
    rewrite (H _ _ _ Hx H').
    apply Hnext_false; auto.
  Qed.

End bbi_sequence.

